#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "which_fetchable",
    "Fetchable",
    "FetchableAtom",
    "FetchableSet",
    "FetchableSeries",
    "Downloadable",
    "DownloadableFile",
    "DownloadableSet",
]

import typing
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Optional,
    Union,
)

from neptune.common.warnings import NeptuneUnsupportedType
from neptune.internal.backends.api_model import (
    Attribute,
    AttributeType,
)
from neptune.internal.container_type import ContainerType
from neptune.internal.utils import verify_type

from neptune_fetcher.attributes import (
    Attr,
    Boolean,
    Datetime,
    Float,
    FloatSeries,
    Integer,
    Series,
    Set,
    String,
    StringSeries,
    StringSet,
)

if TYPE_CHECKING:
    from pandas import DataFrame

    from neptune_fetcher.custom_backend import CustomBackend


ATOMS = {
    AttributeType.INT,
    AttributeType.FLOAT,
    AttributeType.BOOL,
    AttributeType.DATETIME,
    AttributeType.STRING,
    AttributeType.GIT_REF,
    AttributeType.RUN_STATE,
    AttributeType.ARTIFACT,
}
SERIES = {AttributeType.FLOAT_SERIES, AttributeType.STRING_SERIES}


class Fetchable(ABC):
    def __init__(
        self,
        attribute: Attribute,
        backend: "CustomBackend",
        container_id: str,
        cache: Dict[str, Union[Attr, Series, Set]],
    ) -> None:
        self._attribute = attribute
        self._backend = backend
        self._container_id = container_id
        self._cache = cache

    @abstractmethod
    def fetch(self):
        ...


class FetchableAtom(Fetchable):
    def fetch(self):
        if self._attribute.path in self._cache:
            # from cache
            return self._cache[self._attribute.path].val
        if self._attribute.type == AttributeType.STRING:
            attr = String(self._attribute.type)
        elif self._attribute.type == AttributeType.INT:
            attr = Integer(self._attribute.type)
        elif self._attribute.type == AttributeType.BOOL:
            attr = Boolean(self._attribute.type)
        elif self._attribute.type == AttributeType.DATETIME:
            attr = Datetime(self._attribute.type)
        elif self._attribute.type == AttributeType.FLOAT:
            attr = Float(self._attribute.type)
        else:
            raise NeptuneUnsupportedType()
        attr.val = attr.fetch(self._backend, self._container_id, ContainerType.RUN, [self._attribute.path])
        self._cache[self._attribute.path] = attr
        return attr.val


class FetchableSeries(Fetchable):
    def fetch(self):
        raise NeptuneUnsupportedType()

    def fetch_values(self, *, include_timestamp: bool = True) -> "DataFrame":
        if self._attribute.path in self._cache:
            # from cache
            return self._cache[self._attribute.path].values
        if self._attribute.type == AttributeType.FLOAT_SERIES:
            series = FloatSeries()
        elif self._attribute.type == AttributeType.STRING_SERIES:
            series = StringSeries()
        else:
            raise NeptuneUnsupportedType()
        series.values = series.fetch_values(
            self._backend, self._container_id, ContainerType.RUN, [self._attribute.path], include_timestamp
        )
        self._cache[self._attribute.path] = series
        return series.values

    def fetch_last(self) -> Union[str, float]:
        if self._attribute.type == AttributeType.FLOAT_SERIES:
            series = FloatSeries()
        elif self._attribute.type == AttributeType.STRING_SERIES:
            series = StringSeries()
        else:
            raise NeptuneUnsupportedType()
        series.last = series.fetch_last(self._backend, self._container_id, ContainerType.RUN, self._attribute.path)
        self._cache[self._attribute.path] = series
        return series.last


class FetchableSet(Fetchable):
    def fetch(self) -> typing.Set[str]:
        if self._attribute.path in self._cache:
            # from cache
            return self._cache[self._attribute.path].values
        if self._attribute.type == AttributeType.STRING_SET:
            s = StringSet()
        else:
            raise NeptuneUnsupportedType()

        s.values = s.fetch(self._backend, self._container_id, ContainerType.RUN, self._attribute.path)
        self._cache[self._attribute.path] = s
        return s.values


class Downloadable(Fetchable, ABC):
    def fetch(self):
        raise NeptuneUnsupportedType("Files and FileSets do not support `fetch` operation")

    @abstractmethod
    def download(self, destination: Optional[str] = None) -> None:
        ...


class DownloadableFile(Downloadable):
    def download(self, destination: Optional[str] = None) -> None:
        verify_type("destination", destination, (str, type(None)))
        self._backend.download_file(self._container_id, ContainerType.RUN, [self._attribute.path], destination)

    def fetch_extension(self) -> str:
        val = self._backend.get_file_attribute(self._container_id, ContainerType.RUN, [self._attribute.path])
        return val.ext


class DownloadableSet(Downloadable):
    def download(self, destination: Optional[str] = None) -> None:
        verify_type("destination", destination, (str, type(None)))
        self._backend.download_file_set(self._container_id, ContainerType.RUN, [self._attribute.path], destination)


def which_fetchable(attribute: Attribute, *args: Any, **kwargs: Any) -> Fetchable:
    if attribute.type in ATOMS:
        return FetchableAtom(attribute, *args, **kwargs)
    elif attribute.type in SERIES:
        return FetchableSeries(attribute, *args, **kwargs)
    elif attribute.type == AttributeType.STRING_SET:
        return FetchableSet(attribute, *args, **kwargs)
    elif attribute.type == AttributeType.FILE:
        return DownloadableFile(attribute, *args, **kwargs)
    elif attribute.type == AttributeType.FILE_SET:
        return DownloadableSet(attribute, *args, **kwargs)
    raise NeptuneUnsupportedType(f"Unsupported type: {attribute.type}")



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["initialize"]

import functools
import os
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Tuple,
    Type,
)

from neptune_experimental.env import NEPTUNE_SAFETY_MODE
from neptune_experimental.utils import override_method

if TYPE_CHECKING:
    from neptune.metadata_containers import MetadataContainer


def initialize() -> None:
    override_run()
    override_project()
    override_model_version()
    override_model()
    override_handler()


def override_common_operations(obj: Type["MetadataContainer"]) -> None:
    from neptune.handler import Handler

    override_method(obj=obj, method="__getitem__", method_factory=safe_function(Handler(None, None)))
    override_method(obj=obj, method="__setitem__", method_factory=safe_function())
    override_method(obj=obj, method="__delitem__", method_factory=safe_function())
    override_method(obj=obj, method="assign", method_factory=safe_function())
    override_method(obj=obj, method="fetch", method_factory=safe_function({}))
    override_method(obj=obj, method="ping", method_factory=safe_function())
    override_method(obj=obj, method="start", method_factory=safe_function())
    override_method(obj=obj, method="stop", method_factory=safe_function())
    override_method(obj=obj, method="get_state", method_factory=safe_function())
    override_method(obj=obj, method="get_structure", method_factory=safe_function({}))
    override_method(obj=obj, method="print_structure", method_factory=safe_function())
    override_method(obj=obj, method="define", method_factory=safe_function())
    override_method(obj=obj, method="get_attribute", method_factory=safe_function())
    override_method(obj=obj, method="set_attribute", method_factory=safe_function())
    override_method(obj=obj, method="exists", method_factory=safe_function(False))
    override_method(obj=obj, method="pop", method_factory=safe_function())
    override_method(obj=obj, method="wait", method_factory=safe_function())
    override_method(obj=obj, method="sync", method_factory=safe_function())
    override_method(obj=obj, method="get_root_object", method_factory=safe_function())
    override_method(obj=obj, method="get_url", method_factory=safe_function())


def override_model() -> None:
    from neptune import Model

    override_common_operations(obj=Model)
    override_method(obj=Model, method="fetch_model_versions_table", method_factory=safe_function())


def override_model_version() -> None:
    from neptune import ModelVersion

    override_common_operations(obj=ModelVersion)
    override_method(obj=ModelVersion, method="change_stage", method_factory=safe_function())


def override_project() -> None:
    from neptune import Project

    override_common_operations(obj=Project)
    override_method(obj=Project, method="fetch_runs_table", method_factory=safe_function())
    override_method(obj=Project, method="fetch_models_table", method_factory=safe_function())


def override_run() -> None:
    from neptune import Run

    override_common_operations(obj=Run)
    override_method(obj=Run, method="monitoring_namespace", method_factory=safe_function())


def override_handler() -> None:
    from neptune.handler import Handler

    override_method(obj=Handler, method="__getitem__", method_factory=safe_function())
    override_method(obj=Handler, method="__setitem__", method_factory=safe_function())
    override_method(obj=Handler, method="__getattr__", method_factory=safe_function())
    override_method(obj=Handler, method="get_root_object", method_factory=safe_function())
    override_method(obj=Handler, method="assign", method_factory=safe_function())
    override_method(obj=Handler, method="upload", method_factory=safe_function())
    override_method(obj=Handler, method="upload_files", method_factory=safe_function())
    override_method(obj=Handler, method="log", method_factory=safe_function())
    override_method(obj=Handler, method="append", method_factory=safe_function())
    override_method(obj=Handler, method="extend", method_factory=safe_function())
    override_method(obj=Handler, method="add", method_factory=safe_function())
    override_method(obj=Handler, method="pop", method_factory=safe_function())
    override_method(obj=Handler, method="remove", method_factory=safe_function())
    override_method(obj=Handler, method="clear", method_factory=safe_function())
    override_method(obj=Handler, method="delete_files", method_factory=safe_function())
    override_method(obj=Handler, method="download", method_factory=safe_function())
    override_method(obj=Handler, method="download_last", method_factory=safe_function())
    override_method(obj=Handler, method="fetch_hash", method_factory=safe_function())
    override_method(obj=Handler, method="fetch_extension", method_factory=safe_function())
    override_method(obj=Handler, method="fetch_files_list", method_factory=safe_function([]))
    override_method(obj=Handler, method="list_fileset_files", method_factory=safe_function([]))
    override_method(obj=Handler, method="track_files", method_factory=safe_function())


def safe_function(default_return_value: Any = None) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    from neptune.common.warnings import warn_once

    safe_mode_enabled = os.getenv(NEPTUNE_SAFETY_MODE, "false").lower() in ("true", "1", "t")

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        def wrapper(*args: Tuple, **kwargs: Dict[str, Any]) -> Any:
            if safe_mode_enabled:
                try:
                    return func(*args, **kwargs)
                except Exception as ex:
                    try:
                        warn_once(f"Exception in method {func}: {ex.__class__.__name__}")
                    except Exception:
                        pass
                    return default_return_value
            else:
                return func(*args, **kwargs)

        return wrapper

    return decorator



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import shutil
import threading
from datetime import datetime
from queue import Queue
from typing import (
    TYPE_CHECKING,
    Any,
    List,
    Optional,
)

from neptune.constants import ASYNC_DIRECTORY
from neptune.internal.backends.neptune_backend import NeptuneBackend
from neptune.internal.container_type import ContainerType
from neptune.internal.id_formats import UniqueId
from neptune.internal.operation import Operation
from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor
from neptune.internal.operation_processors.operation_processor import OperationProcessor
from neptune.internal.operation_processors.operation_storage import get_container_dir

if TYPE_CHECKING:
    from neptune.internal.signals_processing.signals import Signal


_logger = logging.getLogger(__name__)


class PartitionedOperationProcessor(OperationProcessor):
    def __init__(
        self,
        container_id: UniqueId,
        container_type: ContainerType,
        backend: NeptuneBackend,
        lock: threading.RLock,
        queue: "Queue[Signal]",
        batch_size: int,
        sleep_time: float = 5,
        partitions: int = 5,
    ):
        self._data_path = self._init_data_path(container_id, container_type)
        self._partitions = partitions
        self._processors = [
            AsyncOperationProcessor(
                container_id=container_id,
                container_type=container_type,
                backend=backend,
                lock=lock,
                queue=queue,
                sleep_time=sleep_time,
                batch_size=batch_size,
                data_path=self._data_path / f"partition-{partition_id}",
                should_print_logs=False,
            )
            for partition_id in range(partitions)
        ]

    @staticmethod
    def _init_data_path(container_id: "UniqueId", container_type: "ContainerType") -> Any:
        now = datetime.now()
        path_suffix = f"exec-{now.timestamp()}-{now.strftime('%Y-%m-%d_%H.%M.%S.%f')}-{os.getpid()}"
        return get_container_dir(ASYNC_DIRECTORY, container_id, container_type, path_suffix)

    def enqueue_operation(self, op: Operation, *, wait: bool) -> None:
        processor = self._get_operation_processor(op.path)
        processor.enqueue_operation(op, wait=wait)

    def _get_operation_processor(self, path: List[str]) -> OperationProcessor:
        path_hash = hash(tuple(path))
        return self._processors[path_hash % self._partitions]

    def pause(self) -> None:
        for processor in self._processors:
            processor.pause()

    def resume(self) -> None:
        for processor in self._processors:
            processor.resume()

    def wait(self) -> None:
        for processor in self._processors:
            processor.wait()

    def flush(self) -> None:
        for processor in self._processors:
            processor.flush()

    def start(self) -> None:
        # TODO: Handle exceptions
        for processor in self._processors:
            processor.start()

    def stop(self, seconds: Optional[float] = None) -> None:
        # TODO: Handle exceptions
        # TODO: Better stop (async?)
        if seconds is not None:
            seconds /= self._partitions

        for processor in self._processors:
            processor.stop(seconds=seconds)

        if all(processor._queue.is_empty() for processor in self._processors):
            shutil.rmtree(self._data_path, ignore_errors=True)

        if not os.listdir(self._data_path.parent):
            shutil.rmtree(self._data_path.parent, ignore_errors=True)

    def close(self) -> None:
        for processor in self._processors:
            processor.close()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "Attr",
    "Boolean",
    "Datetime",
    "Float",
    "FloatSeries",
    "Integer",
    "Set",
    "Series",
    "String",
    "StringSeries",
    "StringSet",
]

import typing
from abc import ABC
from dataclasses import dataclass
from datetime import datetime
from typing import (
    Dict,
    Generic,
    Optional,
    TypeVar,
    Union,
)

from neptune.attributes.atoms.boolean import Boolean as BooleanAttr
from neptune.attributes.atoms.datetime import Datetime as DatetimeAttr
from neptune.attributes.atoms.float import Float as FloatAttr
from neptune.attributes.atoms.integer import Integer as IntegerAttr
from neptune.attributes.atoms.string import String as StringAttr
from neptune.attributes.series.fetchable_series import Row
from neptune.internal.backends.api_model import (
    AttributeType,
    FloatSeriesValues,
    StringSeriesValues,
)
from neptune.internal.container_type import ContainerType

if typing.TYPE_CHECKING:
    from pandas import DataFrame

    from neptune_fetcher.custom_backend import CustomBackend

T = TypeVar("T")


@dataclass
class Set(Generic[T], ABC):
    values: Optional[typing.Set[T]] = None

    @staticmethod
    def fetch(backend, container_id, container_type, path) -> typing.Set[T]:
        ...


class StringSet(Set[str]):
    @staticmethod
    def fetch(backend, container_id, container_type, path) -> typing.Set[str]:
        return backend.get_string_set_attribute(container_id, container_type, path).values


@dataclass
class Series(ABC, Generic[T]):
    values: Optional["DataFrame"] = None
    last: Optional[T] = None

    def fetch_values(
        self,
        backend: "CustomBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
        include_timestamp: bool = True,
    ) -> "DataFrame":
        import pandas as pd

        limit = 1000
        val = self._fetch_values_from_backend(backend, container_id, container_type, path, 0, limit)
        data = val.values
        offset = limit

        def make_row(entry: Row) -> Dict[str, Union[str, float, datetime]]:
            row: Dict[str, Union[str, float, datetime]] = dict()
            row["step"] = entry.step
            row["value"] = entry.value
            if include_timestamp:
                row["timestamp"] = datetime.fromtimestamp(entry.timestampMillis / 1000)
            return row

        backend.progress_update_handler.pre_series_fetch(val.totalItemCount, limit)
        while offset < val.totalItemCount:
            batch = self._fetch_values_from_backend(backend, container_id, container_type, path, offset, limit)
            data.extend(batch.values)
            offset += limit
            backend.progress_update_handler.on_series_fetch(limit)
        backend.progress_update_handler.post_series_fetch()

        rows = dict((n, make_row(entry)) for (n, entry) in enumerate(data))

        df = pd.DataFrame.from_dict(data=rows, orient="index")
        return df

    @staticmethod
    def _fetch_values_from_backend(
        backend: "CustomBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
        offset: int,
        limit: int,
    ) -> Row:
        ...

    @staticmethod
    def fetch_last(backend: "CustomBackend", container_id: str, container_type: ContainerType, path: str) -> T:
        ...


class FloatSeries(Series[float]):
    @staticmethod
    def _fetch_values_from_backend(
        backend: "CustomBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
        offset: int,
        limit: int,
    ) -> FloatSeriesValues:
        return backend.get_float_series_values(container_id, container_type, path, offset, limit)

    @staticmethod
    def fetch_last(backend: "CustomBackend", container_id: str, container_type: ContainerType, path: str) -> float:
        return backend.get_float_series_attribute(container_id, container_type, [path]).last


class StringSeries(Series[str]):
    @staticmethod
    def _fetch_values_from_backend(
        backend: "CustomBackend",
        container_id: str,
        container_type: ContainerType,
        path: typing.List[str],
        offset: int,
        limit: int,
    ) -> StringSeriesValues:
        return backend.get_string_series_values(container_id, container_type, path, offset, limit)

    @staticmethod
    def fetch_last(backend: "CustomBackend", container_id: str, container_type: ContainerType, path: str) -> str:
        return backend.get_string_series_attribute(container_id, container_type, [path]).last


@dataclass
class Attr(Generic[T], ABC):
    type: AttributeType
    val: Optional[T] = None

    @staticmethod
    def fetch(backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]) -> T:
        ...


class Integer(Attr[int]):
    @staticmethod
    def fetch(
        backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]
    ) -> int:
        return IntegerAttr.getter(
            backend=backend,
            container_id=container_id,
            container_type=container_type,
            path=path,
        )


class Float(Attr[float]):
    @staticmethod
    def fetch(
        backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]
    ) -> float:
        return FloatAttr.getter(
            backend=backend,
            container_id=container_id,
            container_type=container_type,
            path=path,
        )


class String(Attr[str]):
    @staticmethod
    def fetch(
        backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]
    ) -> str:
        return StringAttr.getter(
            backend=backend,
            container_id=container_id,
            container_type=container_type,
            path=path,
        )


class Boolean(Attr[bool]):
    @staticmethod
    def fetch(
        backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]
    ) -> bool:
        return BooleanAttr.getter(
            backend=backend,
            container_id=container_id,
            container_type=container_type,
            path=path,
        )


class Datetime(Attr[datetime]):
    @staticmethod
    def fetch(
        backend: "CustomBackend", container_id: str, container_type: ContainerType, path: typing.List[str]
    ) -> datetime:
        return DatetimeAttr.getter(
            backend=backend,
            container_id=container_id,
            container_type=container_type,
            path=path,
        )



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["initialize"]

import os
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
)

from requests.adapters import (
    DEFAULT_POOLSIZE,
    HTTPAdapter,
)

if TYPE_CHECKING:
    from bravado.requests_client import RequestsClient
    from neptune.internal.operation_processors.operation_processor import OperationProcessor

from neptune_experimental.env import NEPTUNE_ASYNC_PARTITIONS_NUMBER
from neptune_experimental.utils import wrap_method

MAX_POOLSIZE = 128


def initialize() -> None:
    from neptune.internal.backends import hosted_client
    from neptune.internal.operation_processors import factory

    wrap_method(obj=hosted_client, method="_set_pool_size", wrapper=custom_set_pool_size)
    wrap_method(obj=factory, method="build_async_operation_processor", wrapper=custom_build_async_op_processor)


def custom_set_pool_size(http_client: "RequestsClient", original: Callable[["RequestsClient"], None]) -> None:
    partitions = int(os.getenv(NEPTUNE_ASYNC_PARTITIONS_NUMBER, "0"))

    if partitions > DEFAULT_POOLSIZE:
        adapter = HTTPAdapter(pool_connections=partitions, pool_maxsize=min(4 * partitions, MAX_POOLSIZE))
        http_client.session.mount("https://", adapter)
        http_client.session.mount("http://", adapter)


def custom_build_async_op_processor(
    *args: Any, original: Callable[..., "OperationProcessor"], **kwargs: Any
) -> "OperationProcessor":
    from neptune.envs import NEPTUNE_ASYNC_BATCH_SIZE
    from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor

    from neptune_experimental.partitioned_operation_processor import PartitionedOperationProcessor

    batch_size = int(os.environ.get(NEPTUNE_ASYNC_BATCH_SIZE) or "1000")
    partitions = int(os.getenv(NEPTUNE_ASYNC_PARTITIONS_NUMBER, "1"))

    if partitions > 1:
        kwargs["batch_size"] = batch_size
        kwargs["partitions"] = partitions

        return PartitionedOperationProcessor(*args, **kwargs)

    return AsyncOperationProcessor(*args, **kwargs, batch_size=batch_size)



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["initialize"]

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
)

from neptune_experimental.utils import wrap_method

if TYPE_CHECKING:
    from neptune import Run
    from neptune.internal.background_job import BackgroundJob


def initialize() -> None:
    from neptune import Run

    # Monkey patching
    wrap_method(obj=Run, method="__init__", wrapper=init_with_enable_remote_signals)
    wrap_method(obj=Run, method="_get_background_jobs", wrapper=get_background_jobs)


def init_with_enable_remote_signals(self: "Run", *args: Any, original: Callable[..., Any], **kwargs: Any) -> None:
    from neptune.internal.utils import verify_type

    enable_remote_signals = kwargs.pop("enable_remote_signals", None)

    if enable_remote_signals is None:
        self._enable_remote_signals = True  # user did not pass this param in kwargs -> default value
    else:

        verify_type("enable_remote_signals", enable_remote_signals, bool)
        self._enable_remote_signals = enable_remote_signals

    original(self, *args, **kwargs)


def get_background_jobs(self: "Run", original: Callable[..., Any]) -> List["BackgroundJob"]:
    from neptune.internal.websockets.websocket_signals_background_job import WebsocketSignalsBackgroundJob

    background_jobs: List["BackgroundJob"] = original(self)

    if not self._enable_remote_signals:
        # filter-out websocket job
        background_jobs = list(
            filter(
                lambda x: not isinstance(x, WebsocketSignalsBackgroundJob),
                background_jobs,
            )
        )

    return background_jobs



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "ProgressUpdateHandler",
    "NullProgressUpdateHandler",
    "DefaultProgressUpdateHandler",
]

from abc import ABC


class ProgressUpdateHandler(ABC):
    """Abstract base class for progress update handlers.

    Example usage:
    >>> from neptune_fetcher import ReadOnlyProject
    >>> from neptune_fetcher.progress_update_handler import ProgressUpdateHandler

    >>> class CustomProgressUpdateHandler(ProgressUpdateHandler):
    >>>     ...  # overwrite the methods you need

    >>> project = ReadOnlyProject(...)
    >>> project.progress_indicator(CustomProgressUpdateHandler())  # use your custom handler to track progress
    """

    def pre_series_fetch(self, total_series: int, series_limit: int) -> None:
        """Executes before a series is fetched. Use it to set up the tracking up.

        Args:
            total_series: Total number of items in the series.
            series_limit: Limit of items fetched in a single iteration.
        """
        ...

    def pre_runs_table_fetch(self) -> None:
        """Executes before a run table is fetched. Use it to set the tracking up."""
        ...

    def pre_download(self, total_size: int) -> None:
        """Executes before a file or a fileset is downloaded. Use it to set the tracking up.

        Args:
            total_size: Total size of the file.
        """
        ...

    def on_series_fetch(self, step: int) -> None:
        """Executes after every iteration of a series fetch. Use it to update the progress.

        Args:
            step: number of items that were fetched during the iteration."""
        ...

    def on_runs_table_fetch(self, step: int) -> None:
        """Executes after every iteration of a runs table fetch. Use it to update the progress.

        Args:
            step: number of items that were fetched during the iteration."""
        ...

    def post_series_fetch(self) -> None:
        """Executes after the series fetch is completed. Use it to clean the tracking up."""
        ...

    def post_runs_table_fetch(self) -> None:
        """Executes after the run table fetch is completed. Use it to clean the tracking up."""

    def on_download_chunk(self, chunk: int) -> None:
        """Executes after every iteration of a file or fileset download. Use it to update the progress.

        Args:
            chunk: Number of bytes that were fetched during the iteration."""
        ...

    def post_download(self) -> None:
        """Executes after the download is completed. Use it to clean the tracking up."""
        ...


class NullProgressUpdateHandler(ProgressUpdateHandler):
    ...


class DefaultProgressUpdateHandler(ProgressUpdateHandler):
    def pre_series_fetch(self, total_series: int, series_limit: int) -> None:
        from tqdm import tqdm

        self._series_bar = tqdm(total=total_series)
        self._series_bar.update(n=series_limit)

    def pre_runs_table_fetch(self) -> None:
        from tqdm import tqdm

        self._table_bar = tqdm()

    def on_series_fetch(self, step: int) -> None:
        self._series_bar.update(n=step)
        self._series_bar.set_description("Fetching series values")

    def post_series_fetch(self) -> None:
        self._series_bar.close()

    def on_runs_table_fetch(self, step: int) -> None:
        self._table_bar.update(n=step)
        self._table_bar.set_description("Fetching runs")

    def post_runs_table_fetch(self) -> None:
        self._table_bar.close()

    def pre_download(self, total_size: int) -> None:
        from tqdm import tqdm

        self._download_bar = tqdm(total=total_size)

    def on_download_chunk(self, chunk: int) -> None:
        self._download_bar.update(n=chunk)
        self._download_bar.set_description("Downloading file")

    def post_download(self) -> None:
        self._download_bar.close()



#
# Copyright (c) 2023, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = ["OperationErrorProcessor"]

import os
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Set,
)

from neptune_experimental.env import NEPTUNE_SAMPLE_SERIES_STEPS_ERRORS
from neptune_experimental.utils import wrap_method

if TYPE_CHECKING:
    from neptune.common.exceptions import NeptuneException
    from neptune.exceptions import MetadataInconsistency
    from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor


def initialize() -> None:
    from neptune.internal.operation_processors.async_operation_processor import AsyncOperationProcessor

    wrap_method(obj=AsyncOperationProcessor.ConsumerThread, method="__init__", wrapper=custom_init)
    wrap_method(obj=AsyncOperationProcessor.ConsumerThread, method="_handle_errors", wrapper=custom_handle_errors)


class OperationErrorProcessor:
    def __init__(self) -> None:
        self._sampling_enabled = os.getenv(NEPTUNE_SAMPLE_SERIES_STEPS_ERRORS, "false").lower() in ("true", "1", "t")
        self._error_sampling_exp = re.compile(
            r"X-coordinates \(step\) must be strictly increasing for series attribute: (.*)\. Invalid point: (.*)"
        )
        self._logged_steps: Set[str] = set()

    def handle(self, errors: List["NeptuneException"]) -> None:
        from neptune.exceptions import MetadataInconsistency
        from neptune.internal.utils.logger import logger

        for error in errors:
            if self._sampling_enabled and isinstance(error, MetadataInconsistency):
                match_exp = self._error_sampling_exp.match(str(error))
                if match_exp:
                    self._handle_not_increased_error_for_step(error, match_exp.group(2))
                    continue

            logger.error("Error occurred during asynchronous operation processing: %s", str(error))

    def _handle_not_increased_error_for_step(self, error: "MetadataInconsistency", step: str) -> None:
        from neptune.internal.utils.logger import logger

        if step not in self._logged_steps:
            self._logged_steps.add(step)
            logger.error(
                f"Error occurred during asynchronous operation processing: {str(error)}. "
                + f"Suppressing other errors for step: {step}."
            )


def custom_init(
    self: "AsyncOperationProcessor.ConsumerThread", *args: Any, original: Callable[..., Any], **kwargs: Any
) -> None:
    original(self, *args, **kwargs)
    self._errors_processor = OperationErrorProcessor()


def custom_handle_errors(
    self: "AsyncOperationProcessor.ConsumerThread", errors: List["NeptuneException"], *args: Any, **kwargs: Any
) -> None:
    self._errors_processor.handle(errors)
